#pragma once

#include <algorithm>
#include <cmath>
#include <cstdint>
#include <limits>
#include <string>
#include <type_traits>

#include <dune/common/power.hh>

namespace AMDiS
{
  namespace Math
  {
    /// Implementation of the absolute value \f$|a|\f$ of arithmetic types.
    template <class T>
    constexpr T abs(T const& a)
    {
      return  a >= 0 ? a : -a;
    }


    /// Implementation of the square \f$ a^2 \f$ of arithmetic types.
    template <class T>
    constexpr auto sqr(T const& a)
    {
      return a*a;
    }

    /// Implementation of the power \f$ v^p \f$ of arithmetic types `T`.
    template <int p, class T>
    constexpr auto pow(T const& v)
    {
      static_assert( p >= 0, "Exponent p in `pow<p>(v)` should be >= 0," );
      return Dune::Power<p>::eval(v);
    }


    /// Implementation of the minimum of two values \f$ min(a,b)\f$ of any type
    /// supporting the `<` relation.
    /// @{

    template <class T0, class T1>
    constexpr auto min(T0 a, T1 b)
    {
      return a < b ? a : b;
    }

#ifndef AMDIS_HAS_CXX_FOLD_EXPRESSIONS
    template <class T0>
    constexpr T0 min(T0 a)
    {
      return a;
    }

    template <class T0, class... Ts>
    constexpr auto min(T0 a, Ts... ts)
    {
      return min(a, min(ts...));
    }
#else
    template <class T0, class... Ts>
    constexpr auto min(T0 a, Ts... ts)
    {
      std::common_type_t<T0,Ts...> result = a;
      return (result = min(result, ts), ...);
    }
#endif

    /// @}


    /// Implementation of the maximum of two values \f$ max(a,b)\f$ of any type
    /// supporting the `>` relation.
    /// @{
    template <class T0, class T1>
    constexpr auto max(T0 a, T1 b)
    {
      return a < b ? b : a;
    }

#ifndef AMDIS_HAS_CXX_FOLD_EXPRESSIONS
    template <class T0>
    constexpr T0 max(T0 a)
    {
      return a;
    }

    template <class T0, class... Ts>
    constexpr auto max(T0 a, Ts... ts)
    {
      return max(a, max(ts...));
    }
#else
    template <class T0, class... Ts>
    constexpr auto max(T0 a, Ts... ts)
    {
      std::common_type_t<T0,Ts...> result = a;
      return (result = max(result, ts), ...);
    }
#endif

    /// @}

  } // end namespace Math


  template <class T>
  inline void nullify(T& a)
  {
    a = 0;
  }

  inline void nullify(std::string& s)
  {
    s = "";
  }

  template <class T>
  constexpr T threshold = std::numeric_limits<T>::epsilon();


  /// Calculates factorial of i
  constexpr std::uint64_t factorial(std::uint64_t i)
  {
    return i <= 1 ? 1 : i * factorial(i - 1);
  }

  /// check for inf and nan values
  inline bool isNumber(double val)
  {
    return !std::isnan(val) && !std::isinf(val);
  }

} // end namespace AMDiS
