#pragma once

#include <algorithm>

#include <amdis/common/Math.hpp>
#include <amdis/common/Concepts.hpp>
#include <amdis/operations/Basic.hpp>
#include <amdis/operations/Composer.hpp>

namespace AMDiS
{
  namespace Operation
  {
    /** \addtogroup operations
     *  @{
     **/

    /// Functor that represents A+B
    struct Plus
    {
#ifdef AMDIS_HAS_CXX_FOLD_EXPRESSIONS
      template <class... Ts>
      constexpr auto operator()(Ts const&... ts) const
      {
        return (ts + ...);
      }
#else
      template <class T0, class T1>
      constexpr auto operator()(T0 const& t0, T1 const& t1) const
      {
        return t0 + t1;
      }

      template <class T0, class... Ts>
      constexpr auto operator()(T0 const& t0, Ts const&... ts) const
      {
        return t0 + operator()(ts...);
      }
#endif
    };

#ifndef DOXYGEN
    // [0] + g => g
    template <class G>
    struct ComposerBuilder<Plus, Zero, G>
      : ComposerBuilder<Plus, G> {};

    // g + [0] => g
    template <class G>
    struct ComposerBuilder<Plus, G, Zero>
      : ComposerBuilder<Plus, G> {};

    // [0] + [0] => [0]
    template <>
    struct ComposerBuilder<Plus, Zero, Zero>
      : ComposerBuilder<Id, Zero> {};
#endif

    template <class... Int>
    int order(Plus, Int... orders)
    {
      return Math::max(int(orders)...);
    }

    template <std::size_t I>
    auto partial(Plus, index_t<I>)
    {
      static_assert((I < 2), "Derivatives of `Plus` only defined for the binary case.");
      return One{};
    }

    // -------------------------------------------------------------------------

    /// Functor that represents A-B
    struct Minus
    {
      template <class T, class S>
      constexpr auto operator()(T const& lhs, S const& rhs) const
      {
        return lhs - rhs;
      }

      friend int order(Minus, int lhs, int rhs)
      {
        return Math::max(lhs, rhs);
      }

      friend auto partial(Minus, index_t<0>)
      {
        return One{};
      }

      friend auto partial(Minus, index_t<1>)
      {
        return StaticConstant<int,-1>{};
      }
    };

#ifndef DOXYGEN
    // [0] - g => -g
    // template <class G>
    // struct ComposerBuilder<Minus, Zero, G>
    //   : ComposerBuilder<Id, Negate> {};

    // g - [0] => g
    template <class G>
    struct ComposerBuilder<Minus, G, Zero>
      : ComposerBuilder<Minus, G> {};

    // [0] - [0] => [0]
    template <>
    struct ComposerBuilder<Minus, Zero, Zero>
      : ComposerBuilder<Id, Zero> {};
#endif

    // -------------------------------------------------------------------------

    /// Functor that represents A*B
    struct Multiplies
    {
#ifdef AMDIS_HAS_CXX_FOLD_EXPRESSIONS
      template <class... Ts>
      constexpr auto operator()(Ts const&... ts) const
      {
        return (ts * ...);
      }
#else
      template <class T, class S>
      constexpr auto operator()(T const& lhs, S const& rhs) const
      {
        return lhs * rhs;
      }
#endif
    };


#ifndef DOXYGEN
    // [0] * g => [0]
    template <class G>
    struct ComposerBuilder<Multiplies, Zero, G>
      : ComposerBuilder<Id, Zero> {};

    // g * [0] => [0]
    template <class G>
    struct ComposerBuilder<Multiplies, G, Zero>
      : ComposerBuilder<Id, Zero> {};

    // [0] * [0] => [0]
    template <>
    struct ComposerBuilder<Multiplies, Zero, Zero>
      : ComposerBuilder<Id, Zero> {};
#endif


    template <class... Int>
    int order(Multiplies, Int... orders)
    {
      return Plus{}(int(orders)...);
    }

    // only for binary *
    // d_0 (x * y) = y, d_1 (x * y) = x
    template <std::size_t I>
    auto partial(Multiplies, index_t<I>)
    {
      static_assert((I < 2), "Derivatives of `Multiplies` only defined for the binary case.");
      return Arg<1-I>{};
    }

    // -------------------------------------------------------------------------

    /// Functor that represents x^p
    template <int p>
    struct Pow
    {
      static_assert(p > 1, "Exponent in power must be non-negative!");

      template <class T>
      constexpr auto operator()(T const& x) const
      {
        return Math::pow<p>(x);
      }

      friend int order(Pow, int d)
      {
        return p*d;
      }

      friend auto partial(Pow, index_t<0>)
      {
        return compose(Multiplies{}, StaticConstant<int,p>{}, Pow<p-1>{});
      }
    };

    using Sqr = Pow<2>;

    /// \see Pow
    template <>
    struct Pow<1>
        : public Id {};

    /// \see Pow
    template <>
    struct Pow<0>
        : public Zero {};

    /// Functor that represents x^p, \see \ref Pow
    struct Pow_
    {
      Pow_(int p)
        : p_(p)
      {
        assert( p_ >= 0 );
      }

      template <class T>
      auto operator()(T const& x) const
      {
        return std::pow(x, p_);
      }

      friend int order(Pow_ P, int d)
      {
        return P.p_ * d;
      }

      friend auto partial(Pow_ P, index_t<0>)
      {
        return compose(Multiplies{}, Constant<int>{P.p_}, Pow_{P.p_-1});
      }

      int p_;
    };

    // -------------------------------------------------------------------------

    /// Functor that represents A/B
    struct Divides
    {
      template <class T, class S>
      constexpr auto operator()(T const& lhs, S const& rhs) const
      {
        return lhs / rhs;
      }

      // d_0 f(x,y) = 1 / y
      friend auto partial(Divides, index_t<0>)
      {
        return compose(Divides{}, One{}, Arg<1>{});
      }

      // d_1 f(x,y) = (y - x)/y^2
      friend auto partial(Divides, index_t<1>)
      {
        return compose(Divides{}, compose(Minus{}, Arg<1>{}, Arg<0>{}),
                                  compose(Pow<2>{}, Arg<1>{}));
      }
    };

    // -------------------------------------------------------------------------

    /// Functor that represents A-B
    struct Negate
    {
      template <class T>
      constexpr auto operator()(T const& x) const
      {
        return -x;
      }

      friend int order(Negate, int d)
      {
        return d;
      }

      friend auto partial(Negate, index_t<0>)
      {
        return StaticConstant<int,-1>{};
      }
    };

    /** @} **/

  } // end namespace Operation
} // end namespace AMDiS
